/**
  @file app-killer.c

  Application killer program.
  
  Copyright (C) 2004-2005 Nokia Corporation.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#include "app-killer.h"

/* Libosso context */
static osso_context_t *osso = NULL;
/* session bus connection */
static DBusConnection *ses_conn = NULL;
static DBusConnection *sys_conn = NULL;
static GMainLoop *mainloop = NULL;
static gboolean do_init_again = FALSE;
static gboolean restart_base_apps = FALSE;

static void libosso_deinit(void);
static void libosso_init(void);

/**
  Send error reply to backup application.
  @param c D-BUS connection.
  @param m D-BUS message.
*/
static void send_restore_error(DBusConnection *c, DBusMessage *m)
{
    DBusMessage *err = NULL;
    dbus_bool_t rc = FALSE;
    assert(c != NULL && m != NULL);
    err = dbus_message_new_error(m, RESTORE_ERROR, NULL);
    if (err == NULL) {
        ULOG_ERR_F("dbus_message_new_error failed");
        return;
    }
    rc = dbus_connection_send(c, err, NULL);
    if (!rc) {
        ULOG_ERR_F("dbus_connection_send failed");
    }
    dbus_message_unref(err);
}

/**
  Send reply (of success) to the sender of message m.
  @param c D-BUS connection
  @param m D-BUS message to create reply for.
  @return TRUE on success, FALSE on failure.
*/
static gboolean
send_success_reply(DBusConnection *c, DBusMessage *m)
{
    DBusMessage *r = NULL;
    dbus_bool_t rc = FALSE;
    assert(c != NULL && m != NULL);
    r = dbus_message_new_method_return(m);
    if (r == NULL) {
        ULOG_ERR_F("dbus_message_new_method_return failed");
        return FALSE;
    }
    rc = dbus_connection_send(c, r, NULL);
    if (!rc) {
        ULOG_ERR_F("dbus_connection_send failed");
        dbus_message_unref(r);
        return FALSE;
    }
    dbus_message_unref(r);
    return TRUE;
}

/**
  Send error reply to locale change message.
  @param c D-BUS connection.
  @param m D-BUS message.
*/
static void send_locale_error(DBusConnection *c, DBusMessage *m)
{
    DBusMessage *err = NULL;
    dbus_bool_t rc = FALSE;
    assert(c != NULL && m != NULL);
    err = dbus_message_new_error(m, LOCALE_ERROR, NULL);
    if (err == NULL) {
        ULOG_ERR_F("dbus_message_new_error failed");
        return;
    }
    rc = dbus_connection_send(c, err, NULL);
    if (!rc) {
        ULOG_ERR_F("dbus_connection_send failed");
    }
    dbus_message_unref(err);
}

/**
  Send error reply to RFS message.
  @param c D-BUS connection.
  @param m D-BUS message.
  @param e error name.
*/
static void send_rfs_error(DBusConnection *c, DBusMessage *m,
                const char* e)
{
    DBusMessage *err = NULL;
    dbus_bool_t rc = FALSE;
    assert(c != NULL && m != NULL);
    err = dbus_message_new_error(m, e, NULL);
    if (err == NULL) {
        ULOG_ERR_F("dbus_message_new_error failed");
        return;
    }
    rc = dbus_connection_send(c, err, NULL);
    if (!rc) {
        ULOG_ERR_F("dbus_connection_send failed");
    }
    dbus_message_unref(err);
}

/**
  Sends the Exit signal for applications.
  @return TRUE on success, FALSE if the signal was not sent.
*/
static gboolean send_exit_signal()
{
    DBusMessage* m = NULL;
    dbus_bool_t ret = FALSE;
    ULOG_DEBUG_F("entering");
    m = dbus_message_new_signal(AK_BROADCAST_OP, AK_BROADCAST_IF,
                                AK_BROADCAST_EXIT);
    if (m == NULL) {
        ULOG_ERR_F("dbus_message_new_signal failed");
        return FALSE;
    }
    assert(ses_conn != NULL);
    ret = dbus_connection_send(ses_conn, m, NULL);
    if (!ret) {
        ULOG_ERR_F("dbus_connection_send failed");
        dbus_message_unref(m);
        return FALSE;
    }
    dbus_message_unref(m);
    ULOG_DEBUG_F("leaving");
    return TRUE;
}

#if 0
/**
  Executes "af-base-apps stop".
  @return TRUE on success, FALSE on failure.
*/
static gboolean shutdown_base_apps()
{
    static char* const args[] = {AF_BASE_APPS, "stop", NULL};
    int ret = exec_prog(AF_BASE_APPS, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Executes "af-base-apps start".
  @return TRUE on success, FALSE on failure.
*/
static gboolean start_base_apps()
{
    static char* const args[] = {AF_BASE_APPS, "start", NULL};
    int ret = exec_prog(AF_BASE_APPS, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Executes "dbus-sessionbus.sh stop".
  @return TRUE on success, FALSE on failure.
*/
static gboolean shutdown_dbus()
{
    static char* const args[] = {DBUS_SESSIONBUS_SCRIPT, "stop", NULL};
    int ret = exec_prog(DBUS_SESSIONBUS_SCRIPT, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Executes "dbus-sessionbus.sh start".
  @return TRUE on success, FALSE on failure.
*/
static gboolean start_dbus()
{
    static char* const args[] = {DBUS_SESSIONBUS_SCRIPT, "start", NULL};
    int ret = exec_prog(DBUS_SESSIONBUS_SCRIPT, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Executes "gconf-daemon.sh stop".
  @return TRUE on success, FALSE on failure.
*/
static gboolean shutdown_gconfd()
{
    static char* const args[] = {GCONF_DAEMON_SCRIPT, "stop", NULL};
    int ret = exec_prog(GCONF_DAEMON_SCRIPT, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}
#endif

/**
  Executes script handling RFS things.
  @return TRUE on success, FALSE on failure.
*/
static gboolean do_rfs()
{
    static char* const args[] = {AK_RFS_SCRIPT, NULL};
    int ret = exec_prog(AK_RFS_SCRIPT, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Executes script handling Locale change things.
  @return TRUE on success, FALSE on failure.
*/
static gboolean do_locale_change()
{
    static char* const args[] = {AK_LOCALE_SCRIPT, NULL};
    int ret = exec_prog(AK_LOCALE_SCRIPT, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Executes script handling Restore things.
  @return TRUE on success, FALSE on failure.
*/
static gboolean do_restore()
{
    static char* const args[] = {AK_RESTORE_SCRIPT, NULL};
    int ret = exec_prog(AK_RESTORE_SCRIPT, args);
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

/**
  Changes to messagebus user.
  @return TRUE on success, FALSE on failure.
*/
static gboolean switch_to_messagebus()
{
    struct passwd *pwd = NULL;
    int ret = -1, uid = -1;

    pwd = getpwnam(MESSAGEBUS_USER);
    if (pwd == NULL) {
        ULOG_ERR_F("getpwnam() failed: %s", strerror(errno));
        return FALSE;
    }
    /* init groups */
    ret = initgroups(MESSAGEBUS_USER, pwd->pw_gid);
    if (ret < 0) {
        ULOG_ERR_F("initgroups() failed: %s", strerror(errno));
        return FALSE;
    }
    uid = pwd->pw_uid;

    if (geteuid() == uid) {
        ULOG_WARN_F("euid is already %d", uid);
    }
    if (seteuid(uid) < 0) {
        ULOG_ERR_F("couldn't change to euid %d", uid);
        return FALSE;
    }
    return TRUE;
}

/**
  Changes to root user.
  @return TRUE on success, FALSE on failure.
*/
static gboolean switch_to_root()
{
    struct passwd *pwd = NULL;
    int ret = -1;

    if (geteuid() == ROOT_UID) {
        ULOG_WARN_F("euid is already %d", ROOT_UID);
    }
    if (seteuid(ROOT_UID) < 0) {
        ULOG_ERR_F("couldn't change to euid %d", ROOT_UID);
        return FALSE;
    }

    pwd = getpwnam(ROOT_USER);
    if (pwd == NULL) {
        ULOG_ERR_F("getpwnam() failed: %s", strerror(errno));
        return FALSE;
    }
    /* init groups */
    ret = initgroups(ROOT_USER, pwd->pw_gid);
    if (ret < 0) {
        ULOG_ERR_F("initgroups() failed: %s", strerror(errno));
        return FALSE;
    }
    return TRUE;
}

/**
  Executes "gconf-daemon.sh start" as messagebus user.
  @return TRUE on success, FALSE on failure.
*/
static gboolean start_gconfd()
{
    int ret = -1;
    static char* const args[] = {GCONF_DAEMON_SCRIPT, "start", NULL};
    if (!switch_to_messagebus()) {
        ULOG_ERR_F("could not change to messagebus user");
        return FALSE;
    }
    ret = exec_prog(GCONF_DAEMON_SCRIPT, args);
    if (!switch_to_root()) {
        ULOG_ERR_F("could not change back to root");
    }
    if (ret) {
        return FALSE;
    } else {
        return TRUE;
    }
}

#if 0
/**
  Message handler callback function for shutting down applications
  for the backup operation.

  @param c D-BUS connection.
  @param m D-BUS message.
  @param data User data passed to the function.
  @return OSSO_OK
*/
static DBusHandlerResult backup_handler(DBusConnection *c,
                                 DBusMessage *m, void *data)
{
    DBusMessage *reply = NULL; 
    dbus_bool_t rc = FALSE;
    ULOG_DEBUG_F("entered");
    if (!send_exit_signal()) {
        ULOG_ERR_F("could not send the exit signal");
        send_backup_error(c, m);
        return OSSO_OK;
    }
    if (!shutdown_base_apps()) {
        ULOG_ERR_F("could not shut down the base apps");
        send_backup_error(c, m);
        return OSSO_OK;
    }

    /* send reply to Backup */
    reply = dbus_message_new_method_return(m);
    if (reply == NULL) {
        ULOG_ERR_F("dbus_message_new_method_return failed");
        return OSSO_OK;
    }
    rc = dbus_connection_send(c, reply, NULL);
    if (!rc) {
        ULOG_ERR_F("dbus_connection_send failed");
    }
    dbus_message_unref(reply);
    return OSSO_OK;
}
#endif

/**
  Sends a window state saving message to Task Navigator.
  Blocks until gets a reply or timeouts.
  @return TRUE on success, FALSE on error.
*/
static gboolean send_save_msg()
{
    DBusMessage *m = NULL;
    DBusMessage *reply = NULL;
    DBusError err;
    assert(ses_conn != NULL);

    m = dbus_message_new_method_call(TN_SVC, TN_OP, TN_IF, TN_SAVE_METHOD);
    if (m == NULL) {
        ULOG_ERR_F("could not create message");
        return FALSE;
    }
    dbus_error_init(&err);
    reply = dbus_connection_send_with_reply_and_block(ses_conn,
                    m, MSG_TIMEOUT, &err);
    if (reply == NULL) {
        if (dbus_error_has_name(&err, DBUS_ERROR_NO_REPLY)) {
            ULOG_ERR_F("no reply from TN received (in time)");
        } else {
            ULOG_ERR_F("error when sending message");
        }
        dbus_message_unref(m);
        return FALSE;
    }
    dbus_message_unref(m);
    return TRUE;
}

/**
  Sends a window restoring message to Task Navigator.
  @return TRUE on success, FALSE on error.
*/
static gboolean send_restore_msg()
{
    DBusMessage *m = NULL;
    DBusMessage *reply = NULL;
    DBusError err;
    assert(ses_conn != NULL);

    m = dbus_message_new_method_call(TN_SVC, TN_OP, TN_IF,
                    TN_RESTORE_METHOD);
    if (m == NULL) {
        ULOG_ERR_F("could not create message");
        return FALSE;
    }
    dbus_error_init(&err);
    reply = dbus_connection_send_with_reply_and_block(ses_conn,
                    m, MSG_TIMEOUT, &err);
    if (reply == NULL) {
        if (dbus_error_has_name(&err, DBUS_ERROR_NO_REPLY)) {
            ULOG_ERR_F("no reply from TN received (in time)");
        } else {
            ULOG_ERR_F("error when sending message");
        }
        dbus_message_unref(m);
        return FALSE;
    }
    dbus_message_unref(m);
    return TRUE;
}

#if 0
/**
  Message handler callback function for restarting some applications
  after the backup operation.

  @param c D-BUS connection.
  @param m D-BUS message.
  @param data User data passed to the function.
  @return OSSO_OK
*/
static DBusHandlerResult backup_end_handler(DBusConnection *c,
                        DBusMessage *m, void *data)
{
    DBusMessage *reply = NULL; 
    dbus_bool_t rc = FALSE;
    ULOG_DEBUG("entered backup_end_handler()");
    /* TODO: restart applications */

    /* send reply to Backup */
    reply = dbus_message_new_method_return(m);
    if (reply == NULL) {
        ULOG_CRIT("dbus_message_new_method_return failed");
        exit(1);
    }
    rc = dbus_connection_send(c, reply, NULL);
    if (!rc) {
        ULOG_CRIT("dbus_connection_send failed");
        exit(1);
    }
    dbus_message_unref(reply);
    return OSSO_OK;
}
#endif

	
#if 0
/**
  Message handler callback function for the low memory action.
  This function broadcasts a D-BUS message that suggests background
  applications to exit.

  @param iface D-BUS interface.
  @param method D-BUS method.
  @param args Method arguments.
  @param data User data passed to the function.
  @param retval Return value for Libosso in case of error.
  @return OSSO_OK
*/
static
gint lowmem_handler(const gchar *iface, const char *method,
                    GArray *args, gpointer data, osso_rpc_t *retval)
{
    ULOG_DEBUG("entered lowmem_handler()");
    ULOG_DEBUG("method: %s", method);
    send_exitif_signal();
    return OSSO_OK;
}
#endif


/**
  Reads and sets locale variables from the locale file.
*/
static void read_locale()
{
    FILE *f = NULL;
    static char buf[MAX_LINE];
    static char var[MAX_LINE];
    static char val[MAX_LINE];

    f = fopen(OSSO_LOCALE_FILE, "r");
    if (f == NULL) {
        ULOG_ERR("fopen() failed: %s", strerror(errno));
        return;
    }
    while (fgets(buf, MAX_LINE, f) != NULL) {
        char *s = strstr(buf, "export ");
        if (s != NULL) {
            int ret = -1;
            s += 7;  /* jump over the "export " */
            ret = sscanf(s, "%[^=]=%s", var, val);
            if (ret != 2) {
                ULOG_ERR_F("could not parse line: '%s'", buf);
                fclose(f);
                return;
            }
            /* TODO: possible double quotes should be handled */
            ret = snprintf(buf, MAX_LINE, "%s=%s", var, val);
            if (ret >= MAX_LINE) {
                ULOG_ERR_F("snprintf() had to truncate");
                /* it makes no sense to continue */
                if (fclose(f) == EOF) {
                    ULOG_ERR_F("fclose() failed: %s", strerror(errno));
                }
                return;
            }
            ULOG_DEBUG_F("giving string '%s' to putenv()", buf);
            if (putenv(buf) < 0) {
                ULOG_ERR_F("putenv() failed");
                /* it makes no sense to continue */
                if (fclose(f) == EOF) {
                    ULOG_ERR_F("fclose() failed: %s", strerror(errno));
                }
                return;
            }
        }
    }
    if (fclose(f) == EOF) {
        ULOG_ERR_F("fclose() failed: %s", strerror(errno));
    }
}

/**
  Message handler callback function for changing locale.
  This function broadcasts a D-BUS message that suggests 
  applications to exit and shuts down the base applications and
  shuts down the session bus. 
  After that, it waits for a while and then re-starts the session
  bus and base applications. Lastly, this function causes the AK
  to exit (to allow use of a shell script and simpler C code); it
  is restarted on demand by the D-BUS daemon.

  @param c D-BUS connection.
  @param m D-BUS message.
  @param data User data passed to the function.
  @return Does not return.
*/
static DBusHandlerResult 
locale_handler(DBusConnection *c, DBusMessage *m, void *data)
{
    ULOG_DEBUG_F("entered");
    ULOG_DEBUG_F("method: %s", dbus_message_get_member(m)); 

    if (!send_save_msg()) {
        ULOG_ERR_F("could not send save message "
                        "to (or no reply from) TN");
        send_locale_error(c, m);
	exit(0);
    }
    if (!send_exit_signal()) {
        ULOG_ERR_F("could not send the exit signal");
        send_locale_error(c, m);
	exit(0);
    }
    if (!do_locale_change()) {
        ULOG_ERR_F("locale change operation wasn't successful");
	/* no point to send error, no-one receives it */
	exit(0);
    }
#if 0
    /* this causes exiting from the main loop when this function
       returns and Disconnect signal is handled */
    if (!shutdown_dbus()) {
        ULOG_ERR_F("could not shut down the session bus");
        /* TODO: We should probably reboot here */
    }
    /* read new locale variables to the environment */
    read_locale();
    /* restart base apps (and notify TN) outside main loop
       (session bus is restarted before base apps in signal handler) */
    restart_base_apps = TRUE;
    return DBUS_HANDLER_RESULT_HANDLED;
#endif
    exit(0);
}


/**
  Message handler callback function for shutting down applications
  because of the restore operation.
  This function broadcasts a D-BUS message that suggests 
  applications to exit and shuts down the base applications and
  kills the GConf daemon.

  @param c D-BUS connection.
  @param m D-BUS message.
  @param data User data passed to the function.
  @return DBUS_HANDLER_RESULT_HANDLED
*/
static DBusHandlerResult 
restore_handler(DBusConnection *c, DBusMessage *m, void *data)
{
    ULOG_DEBUG_F("entered");
    ULOG_DEBUG_F("method: %s", dbus_message_get_member(m)); 

    if (!send_exit_signal()) {
        ULOG_ERR_F("could not send the exit signal");
        send_restore_error(c, m);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
    if (!do_restore()) {
        ULOG_ERR_F("could not shutdown programs for restore operation");
        send_restore_error(c, m);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
#if 0
    if (!shutdown_base_apps()) {
        ULOG_ERR_F("could not shut down the base apps");
        send_restore_error(c, m);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
    if (!shutdown_gconfd()) {
        ULOG_ERR_F("could not shut down the GConf daemon");
        send_restore_error(c, m);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
#endif
    if (!send_success_reply(c, m)) {
        ULOG_ERR_F("could not send reply to Backup");
        /* most probably this also fails in this situation... */
        send_restore_error(c, m);
    }
    return DBUS_HANDLER_RESULT_HANDLED;
}

/**
  Message handler callback function for shutting down applications
  because of the RFS operation.
  This function broadcasts a D-BUS message that suggests 
  applications to exit, kills the GConf daemon, restarts base
  applications and D-BUS session bus. Lastly, the AK exits (to
  enable simpler C code and use of a shell script).

  @param c D-BUS connection.
  @param m D-BUS message.
  @param data User data passed to the function.
  @return Does not return
*/
static DBusHandlerResult 
rfs_shutdown_handler(DBusConnection *c, DBusMessage *m, void *data)
{
    ULOG_DEBUG_F("entered");
    ULOG_DEBUG_F("method: %s", dbus_message_get_member(m)); 

    if (!send_exit_signal()) {
        ULOG_ERR_F("could not send the exit signal");
        send_rfs_error(c, m, RFS_SHUTDOWN_ERROR);
	exit(0);
    }
    if (!do_rfs()) {
        ULOG_ERR_F("could not shut down and restart programs for RFS");
        send_rfs_error(c, m, RFS_SHUTDOWN_ERROR);
	exit(0);
    }
#if 0
    if (!shutdown_base_apps()) {
        ULOG_ERR_F("could not shut down the base apps");
        send_rfs_error(c, m, RFS_SHUTDOWN_ERROR);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
    if (!shutdown_gconfd()) {
        ULOG_ERR_F("could not shut down the GConf daemon");
        send_rfs_error(c, m, RFS_SHUTDOWN_ERROR);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
    if (!send_success_reply(c, m)) {
        ULOG_ERR_F("could not send reply");
        /* most probably this also fails in this situation... */
        send_rfs_error(c, m, RFS_SHUTDOWN_ERROR);
    }
    return DBUS_HANDLER_RESULT_HANDLED;
#endif
    exit(0);
}

#if 0
/**
  Message handler callback function for restarting applications
  because of the RFS operation.
  This function does "af-base-apps start".

  @param c D-BUS connection.
  @param m D-BUS message.
  @param data User data passed to the function.
  @return DBUS_HANDLER_RESULT_HANDLED
*/
static DBusHandlerResult 
rfs_restart_handler(DBusConnection *c, DBusMessage *m, void *data)
{
    ULOG_DEBUG_F("entered");
    ULOG_DEBUG_F("method: %s", dbus_message_get_member(m)); 

    /* this is needed to get the PID file */
    if (!start_gconfd()) {
        ULOG_ERR_F("could not restart the Gconf daemon");
        send_rfs_error(c, m, RFS_RESTART_ERROR);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
    if (!start_base_apps()) {
        ULOG_ERR_F("could not restart the base apps");
        send_rfs_error(c, m, RFS_RESTART_ERROR);
        return DBUS_HANDLER_RESULT_HANDLED;
    }
    if (!send_success_reply(c, m)) {
        ULOG_ERR_F("could not send reply");
        /* most probably this also fails in this situation... */
        send_rfs_error(c, m, RFS_RESTART_ERROR);
    }
    return DBUS_HANDLER_RESULT_HANDLED;
}
#endif

/**
  System bus message filter.
*/
static
DBusHandlerResult sysbus_filter(DBusConnection* c,
                                 DBusMessage* m,
                                 void* user_data)
{
#if 0
    printf("sysbus_filter, if: %s\n", dbus_message_get_interface(m));
    printf("sysbus_filter, member: %s\n", dbus_message_get_member(m));
    printf("sysbus_filter, path: %s\n", dbus_message_get_path(m));
    printf("sysbus_filter, sender: %s\n", dbus_message_get_sender(m));
#endif

    if (dbus_message_is_signal(m, DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL,
                               "Disconnected")) {
        ULOG_CRIT("D-BUS system bus disconnected");
	exit(1);
    }
    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

/**
  Session bus message filter.
*/
static
DBusHandlerResult sesbus_filter(DBusConnection* c,
                                 DBusMessage* m,
                                 void* user_data)
{
#if 0
    printf("sesbus_filter, if: %s\n", dbus_message_get_interface(m));
    printf("sesbus_filter, member: %s\n", dbus_message_get_member(m));
    printf("sesbus_filter, path: %s\n", dbus_message_get_path(m));
    printf("sesbus_filter, sender: %s\n", dbus_message_get_sender(m));
#endif

    if (dbus_message_is_signal(m, DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL,
                               "Disconnected")) {
        ULOG_DEBUG("D-BUS session bus disconnected");
#if 0
        libosso_deinit();

	ULOG_INFO("restarting D-BUS session bus");
        if (!start_dbus()) {
	    ULOG_CRIT("failed to restart D-BUS session bus");
            /* TODO: reboot in this case! */
        }

	/* Reiniting Libosso needs to be done outside main loop */
	assert(mainloop != NULL);
	do_init_again = TRUE;
	g_main_loop_quit(mainloop);
#endif
	exit(1);
    }
    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

#if 0
/**
  Sends the Exit-if-possible signal for applications.
*/
static void send_exitif_signal()
{
    DBusMessage* m = NULL;
    dbus_bool_t ret = FALSE;
    ULOG_DEBUG("entering send_exitif_signal");
    m = dbus_message_new_signal(AK_BROADCAST_OP, AK_BROADCAST_IF,
                                AK_BROADCAST_EXITIF);
    if (m == NULL) {
        ULOG_CRIT("dbus_message_new_signal failed");
	exit(1);
    }
    assert(ses_conn != NULL);
    ret = dbus_connection_send(ses_conn, m, NULL);
    if (!ret) {
        ULOG_CRIT("dbus_connection_send failed");
        exit(1);
    }
    ULOG_DEBUG("leaving send_exitif_signal");
}
#endif


/**
  De-initialise Libosso.
*/
static void libosso_deinit()
{
    ULOG_DEBUG("entering libosso_deinit");
    assert(osso != NULL);
    osso_deinitialize(osso);
    osso = NULL;
    ses_conn = NULL;
    assert(osso == NULL && ses_conn == NULL);
}

/**
  Initialise Libosso and register object paths etc.
*/
static void libosso_init()
{
    static DBusObjectPathVTable bu_vt = { 
        .message_function = locale_handler,
        .unregister_function = NULL
    };
    dbus_bool_t rc = FALSE;
    ULOG_DEBUG("entering libosso_init");
    assert(osso == NULL); /* don't initialise Libosso twice in a row */
    assert(ses_conn == NULL);

    osso = osso_initialize(APPL_DBUS_NAME, APPL_VERSION, TRUE, NULL);
    if (osso == NULL) {
        ULOG_CRIT("Libosso initialisation failed");
        exit(1);
    }

    ses_conn = (DBusConnection*) osso_get_dbus_connection(osso);
    if (ses_conn == NULL) {
        ULOG_CRIT("osso_get_dbus_connection() failed");
        exit(1);
    }
    /*
    bu_vt.message_function = backup_handler;
    rc = dbus_connection_register_object_path(ses_conn, BU_START_OP, 
                                              &bu_vt, NULL); 
    if (!rc) {
        ULOG_CRIT("Failed to register backup start callback");
	exit(1);
    }
    */
    
    /*
    bu_vt.message_function = backup_end_handler;
    rc = dbus_connection_register_object_path(ses_conn, BU_END_OP, 
                            &bu_vt, NULL); 
    if (!rc) {
        ULOG_CRIT("Failed to register backup end callback");
	exit(1);
    }
    */

    /*
    ret = osso_rpc_set_cb_f(osso, SVC_NAME, LOWMEM_OP, LOWMEM_IF,
                            lowmem_handler, NULL); 
    if (ret != OSSO_OK) {
        ULOG_CRIT("Failed to register low memory callback");
	exit(1);
    }
    */

    /*
    ret = osso_rpc_set_cb_f(osso, SVC_NAME, LOCALE_OP, LOCALE_IF,
                            locale_handler, NULL); 
    if (ret != OSSO_OK) {
        ULOG_CRIT("Failed to register locale change callback");
	exit(1);
    }
    */

    /* set message filters */

    if (!dbus_connection_add_filter(ses_conn, sesbus_filter, NULL, NULL)) {
        ULOG_CRIT("could not register session bus filter function");
        exit(1);
    }
    sys_conn = (DBusConnection*) osso_get_sys_dbus_connection(osso);
    if (sys_conn == NULL) {
        ULOG_CRIT("osso_get_sys_dbus_connection() failed");
        exit(1);
    }
    if (!dbus_connection_add_filter(sys_conn, sysbus_filter,
                            NULL, NULL)) {
        ULOG_CRIT("could not register system bus filter function");
        exit(1);
    }

    /* reg. locale handler */
    bu_vt.message_function = locale_handler;
    rc = dbus_connection_register_object_path(ses_conn, LOCALE_OP,
                    &bu_vt, NULL);
    if (!rc) {
        ULOG_CRIT("could not register locale handler");
        exit(1);
    }
    /* reg. restore handler */
    bu_vt.message_function = restore_handler;
    rc = dbus_connection_register_object_path(ses_conn, RESTORE_OP,
                    &bu_vt, NULL);
    if (!rc) {
        ULOG_CRIT("could not register restore handler");
        exit(1);
    }
    /* reg. RFS shutdown handler */
    bu_vt.message_function = rfs_shutdown_handler;
    rc = dbus_connection_register_object_path(ses_conn, RFS_SHUTDOWN_OP,
                    &bu_vt, NULL);
    if (!rc) {
        ULOG_CRIT("could not register RFS shutdown handler");
        exit(1);
    }
#if 0
    /* reg. RFS restart handler */
    bu_vt.message_function = rfs_restart_handler;
    rc = dbus_connection_register_object_path(ses_conn, RFS_RESTART_OP,
                    &bu_vt, NULL);
    if (!rc) {
        ULOG_CRIT("could not register RFS restart handler");
        exit(1);
    }
#endif
    assert(osso != NULL); 
    assert(ses_conn != NULL);
    assert(sys_conn != NULL);
}

/**
  The main function of app-killer.
  Does initialisations and goes to the Glib main loop.
*/
int main()
{
    char* ses_bus_socket = NULL;
    ULOG_OPEN(APPL_NAME);

    mainloop = g_main_loop_new(NULL, TRUE);
    ULOG_DEBUG("Glib main loop created");

    ses_bus_socket = getenv("DBUS_SESSION_BUS_ADDRESS");
    if (ses_bus_socket == NULL) {
        ULOG_CRIT("DBUS_SESSION_BUS_SOCKET is not defined");
        exit(1);
    }

    /*
init_again:
*/
    libosso_init();
#if 0
    
    if (restart_base_apps) {
        /* restart base apps and ask TN to restore the rest */
        if (!start_base_apps()) {
            ULOG_CRIT("failed to restart base applications");
            /* TODO: reboot should be done here! */
        }
        /* TODO: wait until TN is up */
        if (!send_restore_msg()) {
            ULOG_CRIT("failed to send restore message to TN");
            /* TODO: reboot should be done here! */
        } 
        restart_base_apps = FALSE;
    }
#endif

    ULOG_DEBUG("Going to the main loop");
    g_main_loop_run(mainloop); 
    ULOG_DEBUG("Returned from the main loop");

#if 0
    /* exiting mail loop is needed, since Libosso init cannot be done
      in message handler callback */
    if (do_init_again) {
        do_init_again = FALSE;
        /* maybe trying connecting to the bus or using waitdbus could
           be better here */
	while (!g_file_test(ses_bus_socket, G_FILE_TEST_EXISTS)) {
	    ULOG_DEBUG("Sleeping a second for session bus to"
			    " properly restart");
	    sleep(1);
	}
        goto init_again; 
    }
#endif
    ULOG_WARN("Glib main loop returned without a known reason");
    exit(0);
}

